from typing import List

import aiohttp

from aiodanbooru.models import DanbooruPost


class DanbooruAPI:
    def __init__(self, base_url: str = "https://danbooru.donmai.us"):
        self.base_url = base_url

    async def _get(
        self, session: aiohttp.ClientSession, endpoint: str, params: dict = None
    ) -> dict:
        url = self.base_url + endpoint
        async with session.get(url, params=params) as response:
            response.raise_for_status()
            return await response.json()

    async def _post(
        self, session: aiohttp.ClientSession, endpoint: str, data: dict
    ) -> dict:
        url = self.base_url + endpoint
        async with session.post(url, json=data) as response:
            response.raise_for_status()
            return await response.json()

    async def get_posts(
        self, tags: List[str] = None, limit: int = None
    ) -> List[DanbooruPost]:
        async with aiohttp.ClientSession() as session:
            endpoint = "/posts.json"
            params = {}
            if tags is not None:
                params["tags"] = " ".join(tags)
            if limit is not None:
                params["limit"] = str(limit)
            response = await self._get(session, endpoint, params)
            posts = [DanbooruPost(**post) for post in response]
            return posts

    async def get_random_post(self) -> DanbooruPost:
        async with aiohttp.ClientSession() as session:
            endpoint = "/posts/random.json"
            response = await self._get(session, endpoint)
            post = DanbooruPost(**response)
            return post



import inspect
from typing import Callable, Union, List

from aiodanbooru.models import DanbooruPost


class Filter:
    async def __call__(self, post: DanbooruPost):
        raise NotImplementedError

    def __invert__(self):
        return InvertFilter(self)

    def __and__(self, other):
        return AndFilter(self, other)

    def __or__(self, other):
        return OrFilter(self, other)


class InvertFilter(Filter):
    def __init__(self, base):
        self.base = base

    async def __call__(self, post: DanbooruPost):
        if inspect.iscoroutinefunction(self.base.__call__):
            x = await self.base(post)
        else:
            x = self.base(post)

        return not x


class AndFilter(Filter):
    def __init__(self, base, other):
        self.base = base
        self.other = other

    async def __call__(self, post: DanbooruPost):
        if inspect.iscoroutinefunction(self.base.__call__):
            x = await self.base(post)
        else:
            x = self.base(post)

        # short circuit
        if not x:
            return False

        if inspect.iscoroutinefunction(self.other.__call__):
            y = await self.other(post)
        else:
            y = self.other(post)

        return x and y


class OrFilter(Filter):
    def __init__(self, base, other):
        self.base = base
        self.other = other

    async def __call__(self, post: DanbooruPost):
        if inspect.iscoroutinefunction(self.base.__call__):
            x = await self.base(post)
        else:
            x = self.base(post)

        # short circuit
        if x:
            return True

        if inspect.iscoroutinefunction(self.other.__call__):
            y = await self.other(post)
        else:
            y = self.other(post)

        return x or y


CUSTOM_FILTER_NAME = "CustomFilter"


def create(func: Callable, name: str = None, **kwargs) -> Filter:
    return type(
        name or func.__name__ or CUSTOM_FILTER_NAME,
        (Filter,),
        {"__call__": func, **kwargs},
    )()


# region all_filter
async def all_filter(_, __):
    return True


# noinspection PyShadowingBuiltins
all = create(all_filter)
"""Filter all messages."""


# endregion


# region author filter
# noinspection PyPep8Naming
class author(Filter, set):
    def __init__(self, users: Union[int, List[int]] = None):
        users = [] if users is None else users if isinstance(users, list) else [users]

        super().__init__(int(u) for u in users)

    async def __call__(self, post: DanbooruPost):
        return post.uploader_id in self


# endregion


# region tag filter
# noinspection PyPep8Naming
class tags(Filter, set):
    def __init__(self, *args: Union[str, List[str]]):
        _tags = []
        for arg in args:
            if isinstance(arg, list):
                _tags.extend(arg)
            else:
                _tags.append(arg)

        super().__init__(str(t) for t in _tags)

    async def __call__(self, post: DanbooruPost):
        return any(t in self for t in post.tags)


# endregion



from typing import Optional, List

import aiohttp
from pydantic import BaseModel, Field, validator
from pydantic.networks import HttpUrl


class DanbooruPost(BaseModel):
    id: int
    uploader_id: int
    approver_id: Optional[int]
    tag_string: str
    tag_string_general: str
    tag_string_artist: str
    tag_string_copyright: str
    tag_string_character: str
    tag_string_meta: str
    rating: Optional[str]
    parent_id: Optional[int]
    source: Optional[str]
    md5: Optional[str] = Field(None, description="MD5 hash of the media file")
    file_url: Optional[HttpUrl] = Field(None, description="URL of the media file")
    large_file_url: Optional[HttpUrl] = Field(
        None, description="URL of the large version of the media file"
    )
    preview_file_url: Optional[HttpUrl] = Field(
        None, description="URL of the preview version of the media file"
    )
    file_ext: Optional[str]
    file_size: Optional[int]
    image_width: Optional[int]
    score: int
    fav_count: int
    tag_count_general: int
    tag_count_artist: int
    tag_count_copyright: int
    tag_count_character: int
    tag_count_meta: int
    last_comment_bumped_at: Optional[str]
    last_noted_at: Optional[str]
    has_children: bool
    image_height: Optional[int]
    created_at: str
    updated_at: str

    @validator("source")
    def validate_source(cls, source):
        if source is not None and len(source) < 1:
            raise ValueError("Source must have at least 1 character")
        return source

    class Config:
        extra = "allow"

    extension: Optional[str] = Field(None, alias="file_ext")

    async def get_media(self, use_large: bool = True) -> bytes:
        if not self.file_url and not self.large_file_url and self.source:
            return await self._get_media_from_source()
        url = (
            self.large_file_url if use_large and self.large_file_url else self.file_url
        )
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                response.raise_for_status()
                return await response.read()

    @property
    def tags(self) -> List[str]:
        return self.tag_string.split()

    @property
    def link(self):
        return f"https://danbooru.donmai.us/posts/{self.id}"

    @property
    def media_url(self):
        return (
            self.large_file_url if self.large_file_url else self.file_url or self.source
        )

    def is_video(self) -> bool:
        return self.extension in ["webm", "mp4"]

    def is_image(self) -> bool:
        return self.extension in ["jpg", "jpeg", "png", "gif"]

    @property
    def filename(self):
        return f"{self.md5}.{self.extension}"

    async def _get_media_from_source(self):
        async with aiohttp.ClientSession() as session:
            if self.source.startswith("https://i.pximg.net"):
                url = self.source.replace("i.pximg.net", "i.pixiv.cat")
            elif self.source.startswith("file://"):
                return b""
            else:
                url = self.source
            async with session.get(url) as response:
                response.raise_for_status()
                return await response.read()



import pytest
from aiodanbooru.api import DanbooruAPI, DanbooruPost


@pytest.mark.asyncio
async def test_get_posts():
    api = DanbooruAPI(base_url="https://danbooru.donmai.us")
    tags = ["cat_girl", "solo"]
    limit = 10

    posts = await api.get_posts(tags=tags, limit=limit)

    assert len(posts) <= limit
    for post in posts:
        assert post.tag_string is not None
        assert post.image_width > 0
        assert post.image_height > 0


@pytest.mark.asyncio
async def test_get_media():
    api = DanbooruAPI(base_url="https://danbooru.donmai.us")
    tags = ["cat_girl", "solo"]
    limit = 1

    posts = await api.get_posts(tags=tags, limit=limit)

    assert len(posts) == 1
    post = posts[0]

    # Testing image retrieval
    if post.is_image():
        media_data = await post.get_media()
        assert isinstance(media_data, bytes)
        assert len(media_data) > 0

    # Testing video retrieval
    if post.is_video():
        media_data = await post.get_media(use_large=False)
        assert isinstance(media_data, bytes)
        assert len(media_data) > 0


@pytest.mark.asyncio
async def test_get_random_post():
    api = DanbooruAPI(base_url="https://danbooru.donmai.us")
    random_post = await api.get_random_post()
    assert random_post is not None
    assert isinstance(random_post, DanbooruPost)


@pytest.mark.asyncio
async def test_dispatcher():
    api = DanbooruAPI()

    post: DanbooruPost = (await api.get_posts(limit=1))[0]
    post_id = post.id

    assert post_id is not None


# Running the tests
if __name__ == "__main__":
    pytest.main(["-v"])



import asyncio
from typing import Callable, Union, List

from aiodanbooru.base.filters import Filter
from aiodanbooru.models import DanbooruPost


class Handler(object):
    def __init__(self, func, filters: Union[List[Filter], Filter] = None):
        self.func: Callable = func
        self.filters: Union[List[Filter], Filter] = filters

    def __call__(self, post: DanbooruPost):
        return (
            self.func(post)
            if not asyncio.iscoroutinefunction(self.func)
            else asyncio.create_task(self.func(post))
        )

    async def check(self, post: DanbooruPost) -> bool:
        if self.filters is None:
            return True
        if isinstance(self.filters, list):
            for f in self.filters:
                if not await f(post):
                    return False
            return True
        return await self.filters(post)

    def __repr__(self):
        return f"<Handler func={self.func.__name__} filters={self.filters}>"



import asyncio
from typing import List, Union

from aiodanbooru.api import DanbooruPost, DanbooruAPI
from aiodanbooru.base.filters import Filter
from aiodanbooru.base.handler import Handler

import logging


class Dispatcher:
    def __init__(self):
        self.queue = asyncio.Queue()
        self._handlers: List[Handler] = []
        self.api = DanbooruAPI()
        self.__last_post_id = None

    async def watch_posts(self):
        while True:
            await asyncio.sleep(0.02)
            try:
                post: DanbooruPost = (await self.api.get_posts(limit=1))[0]
            except Exception as e:
                logging.error(e)
                continue
            post_id = post.id
            if self.__last_post_id is None:
                self.__last_post_id = post_id
            elif post_id != self.__last_post_id:
                self.__last_post_id = post_id
                await self.queue.put(post)

    async def handle_events(self):
        while True:
            post = await self.queue.get()
            for handler in self._handlers:
                if await handler.check(post) is False:
                    continue
                if asyncio.iscoroutinefunction(handler.func):
                    await handler.func(post)
                else:
                    handler.func(post)
            self.queue.task_done()

    def new_post(self, filters: Union[List[Filter], Filter] = None):
        def wrapper(func):
            self._handlers.append(Handler(func, filters))
            return func

        return wrapper

    def start(self):
        loop = asyncio.get_event_loop()
        loop.create_task(self.watch_posts())
        loop.create_task(self.handle_events())
        loop.run_forever()


if __name__ == "__main__":
    from aiodanbooru.base import filters

    dispatcher = Dispatcher()

    @dispatcher.new_post(filters.all)
    async def handle_new_post(post_id):
        print(f"New post with ID {post_id} has appeared on the site!")

    dispatcher.start()



from setuptools import setup, find_packages

setup(
    name="aiodanbooru",
    version="1.0.8",
    description="A Python library for interacting with the Danbooru API",
    author="lrdcxdes",
    author_email="lordgrief176@email.com",
    packages=find_packages(exclude=["tests"]),
    install_requires=[
        "aiohttp",
        "pydantic<2.0.0",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    python_requires=">=3.7",
    project_urls={
        "Source": "https://github.com/lrdcxdes/aiodanbooru",
        "Bug Reports": "https://github.com/lrdcxdes/aiodanbooru/issues",
    },
    long_description=open("LONG.rst").read(),
    license="MIT",
)
